use std::{
    borrow::Cow,
    fs::File,
    io::{self, BufWriter, Write},
    path::PathBuf,
};

use ropey::{LineType, Rope};

use crate::{
    change::{ApplyChange, Change},
    editor::EditorState,
    history::History,
};

/// The line type that should be used as the standard line metric in a buffer.
pub const BUFLINE: LineType = LineType::LF_CR;

/// A path for an open text buffer.
///
/// This indicates where the text data of the buffer came from, and
/// where it should be saved to.
#[derive(Debug, Clone, Eq, PartialEq)]
pub enum BufferPath {
    File(PathBuf), // A buffer for a normal file on disk.
    Temp(usize),   // A temporary buffer, with a number ID.
}

/// An open text buffer, currently being edited.
#[derive(Debug, Clone)]
pub struct Buffer {
    pub text: Rope, // The actual text content.
    pub path: BufferPath,
    pub edits_since_saved: i64, // Tracks whether this buffer is out of sync with the last save.
    pub editor_states: Vec<EditorState>,

    history: History,
}

impl Buffer {
    pub fn new(text: Rope, path: BufferPath) -> Buffer {
        Buffer {
            path: path,
            edits_since_saved: 0,
            text: text,
            editor_states: Vec::new(),
            history: History::new(),
        }
    }

    pub fn is_dirty(&self) -> bool {
        self.edits_since_saved != 0
    }

    /// Saves the buffer to disk.
    pub fn save(&mut self) -> io::Result<()> {
        if let BufferPath::File(ref file_path) = self.path {
            let mut f = BufWriter::new(File::create(file_path)?);

            for c in self.text.chunks() {
                f.write(c.as_bytes())?;
            }
            self.edits_since_saved = 0;
        } else {
            return Err(io::Error::new(
                io::ErrorKind::NotFound,
                "Cannot save: buffer has no on-disk file associated with it.",
            ));
        }

        Ok(())
    }

    /// Replaces the given range of chars with the given text.
    ///
    /// The range does not have to be ordered (i.e. the first component can be
    /// greater than the second).
    pub fn edit(&mut self, byte_idx_range: (usize, usize), text: &str) {
        if self.edits_since_saved >= 0 {
            self.edits_since_saved = self.edits_since_saved.wrapping_add(1);
        } else {
            // The undo state matching the last save is inaccessible, so
            // set this to a similarly inaccessible number.
            self.edits_since_saved = i64::MAX;
        }

        // Get the range, properly ordered.
        let (start, end) = if byte_idx_range.0 < byte_idx_range.1 {
            (byte_idx_range.0, byte_idx_range.1)
        } else {
            (byte_idx_range.1, byte_idx_range.0)
        };

        // Build the change.
        let change = Change::from_edit(start, &Cow::from(self.text.slice(start..end)), text);

        // Update editor states.
        for context in self.editor_states.iter_mut() {
            context.apply_change(&change);
        }

        // Apply the edit to the document.
        self.text.apply_change(&change);

        // Update undo stack.
        self.history.push_edit(change);
    }

    /// Un-does the last edit if there is one, and returns the
    /// Change that was inverse-applied, in case it's needed for
    /// further processing.
    ///
    /// Returns None if there is no edit to undo.
    pub fn undo(&mut self) -> Option<&Change> {
        if let Some(change) = self.history.undo() {
            self.edits_since_saved = self.edits_since_saved.wrapping_sub(1);

            // Update editor states.
            for context in self.editor_states.iter_mut() {
                context.apply_change_inverse(change);
            }

            // Apply the edit to the document.
            self.text.apply_change_inverse(change);

            return Some(change);
        } else {
            return None;
        }
    }

    /// Re-does the last edit if there is one, and returns the
    /// Change that was applied, in case it's needed for further
    /// processing.
    ///
    /// Returns None if there is no edit to redo.
    pub fn redo(&mut self) -> Option<&Change> {
        if let Some(change) = self.history.redo() {
            self.edits_since_saved = self.edits_since_saved.wrapping_add(1);

            // Update editor states.
            for context in self.editor_states.iter_mut() {
                context.apply_change(change);
            }

            // Apply the edit to the document.
            self.text.apply_change(change);

            return Some(change);
        } else {
            return None;
        }
    }

    /// Creates a new editor state for the buffer, and returns its index.
    pub fn add_editor_state(&mut self) -> usize {
        self.editor_states.push(EditorState::new());
        return self.editor_states.len() - 1;
    }
}
