use ropey::Rope;

#[derive(Debug, Clone, Copy)]
pub enum Op {
    Retain(usize), // In bytes.
    Replace {
        // These both represent strings, and are byte-index ranges into
        // the Change's `buffer` where their actual string data is
        // stored.  `(0, 0)` can be used to indicate no string data.
        old: (usize, usize),
        new: (usize, usize),
    },
}

impl Op {
    /// The length of the string segment this Op represents *before*
    /// its application.  In bytes.
    pub fn len_pre(&self) -> usize {
        match self {
            Op::Retain(byte_count) => *byte_count,
            Op::Replace { old, .. } => old.1 - old.0,
        }
    }

    /// The length of the string segment this Op represents *after*
    /// its application.  In bytes.
    pub fn len_post(&self) -> usize {
        match self {
            Op::Retain(byte_count) => *byte_count,
            Op::Replace { new, .. } => new.1 - new.0,
        }
    }
}

/// A reversable set of edits treated as an atomic unit.
#[derive(Clone)]
pub struct Change {
    pub ops: Vec<Op>,
    pub buffer: String,
}

impl Change {
    pub fn new() -> Change {
        Change {
            ops: Vec::new(),
            buffer: String::new(),
        }
    }

    /// Creates a Change from a single edit.
    pub fn from_edit(byte_idx: usize, old: &str, new: &str) -> Change {
        let mut buffer = String::new();
        buffer.push_str(old);
        buffer.push_str(new);

        let ops = vec![
            Op::Retain(byte_idx),
            Op::Replace {
                old: (0, old.len()),
                new: (old.len(), old.len() + new.len()),
            },
        ];

        Change {
            ops: ops,
            buffer: buffer,
        }
    }

    /// Creates a Change from a sorted, non-overlapping set of
    /// simultaneous edits.
    ///
    /// Takes an iterator that yields `(byte_index, old_text, replacement_text)`
    /// tuples, representing the edits.  The items are expected to be
    /// yielded in byte-index order, and the byte indices are relative to
    /// the original text--this function does _not_ treat them as a
    /// sequence of progressively applied edits, but rather as a set of
    /// edits applied simultaneously.
    pub fn from_ordered_edit_set<'a, I>(edit_iter: I) -> Change
    where
        I: Iterator<Item = (usize, &'a str, &'a str)> + 'a,
    {
        let mut change = Change::new();
        let mut i = 0;
        let mut len_delta = 0isize;
        for (byte_idx, old, new) in edit_iter {
            let adjusted_byte_idx = (byte_idx as isize + len_delta) as usize;
            let retained = adjusted_byte_idx - i;

            let old_range = (change.buffer.len(), change.buffer.len() + old.len());
            change.buffer.push_str(old);
            let new_range = (change.buffer.len(), change.buffer.len() + new.len());
            change.buffer.push_str(new);

            if retained > 0 {
                change.ops.push(Op::Retain(retained));
            }
            change.ops.push(Op::Replace {
                old: old_range,
                new: new_range,
            });

            i += retained + new.len();
            len_delta += new.len() as isize - old.len() as isize;
        }
        change
    }

    /// Build a Change that is functionally identical to applying
    /// first `self` and then `other` sequentially.
    #[must_use]
    pub fn compose(&self, other: &Change) -> Change {
        use Op::*;

        let mut change = Change::new();

        let mut ops1 = self.ops.iter().copied();
        let mut ops2 = other.ops.iter().copied();
        let mut next_op1 = ops1.next();
        let mut next_op2 = ops2.next();

        // We progress through both change's ops such that their
        // heads stay in sync.  To do this, we have to split ops
        // sometimes, fragmenting the change.
        loop {
            match (next_op1, next_op2) {
                // Done!
                (None, None) => break,

                // One of the changes is empty, so just append the
                // remaining ops from the other one.
                (Some(op), None) => {
                    change.push_op(op, &self.buffer);
                    next_op1 = ops1.next();
                }
                (None, Some(op)) => {
                    change.push_op(op, &other.buffer);
                    next_op2 = ops2.next();
                }

                //----------------
                // Retain, Retain
                (Some(Retain(bc1)), Some(Retain(bc2))) => {
                    if bc1 < bc2 {
                        change.retain(bc1);
                        next_op1 = ops1.next();
                        next_op2 = Some(Retain(bc2 - bc1));
                    } else if bc2 < bc1 {
                        change.retain(bc2);
                        next_op1 = Some(Retain(bc1 - bc2));
                        next_op2 = ops2.next();
                    } else {
                        change.retain(bc1);
                        next_op1 = ops1.next();
                        next_op2 = ops2.next();
                    }
                }

                //-----------------
                // Replace, Retain
                (
                    Some(
                        op1 @ Replace {
                            old: old1,
                            new: new1,
                        },
                    ),
                    Some(Retain(bc2)),
                ) => {
                    if op1.len_post() < bc2 {
                        change.push_op(op1, &self.buffer);
                        next_op1 = ops1.next();
                        next_op2 = Some(Retain(bc2 - op1.len_post()));
                    } else if op1.len_post() > bc2 {
                        let op1a = Op::Replace {
                            old: old1,
                            new: (new1.0, new1.0 + bc2),
                        };
                        let op1b = Op::Replace {
                            old: (0, 0),
                            new: (new1.0 + bc2, new1.1),
                        };
                        change.push_op(op1a, &self.buffer);
                        next_op1 = Some(op1b);
                        next_op2 = ops2.next();
                    } else {
                        change.push_op(op1, &self.buffer);
                        next_op1 = ops1.next();
                        next_op2 = ops2.next();
                    }
                }

                //-----------------
                // Retain, Replace
                (
                    Some(Retain(bc1)),
                    Some(Replace {
                        old: old2,
                        new: new2,
                    }),
                ) => {
                    let op2 = next_op2.unwrap();

                    if bc1 < op2.len_pre() {
                        change.push_op(
                            Replace {
                                old: (old2.0, old2.0 + bc1),
                                new: (0, 0),
                            },
                            &other.buffer,
                        );
                        next_op1 = ops1.next();
                        next_op2 = Some(Replace {
                            old: (old2.0 + bc1, old2.1),
                            new: new2,
                        });
                    } else if bc1 > op2.len_pre() {
                        change.push_op(op2, &other.buffer);
                        next_op1 = Some(Retain(bc1 - op2.len_pre()));
                        next_op2 = ops2.next();
                    } else {
                        change.push_op(op2, &other.buffer);
                        next_op1 = ops1.next();
                        next_op2 = ops2.next();
                    }
                }

                //------------------
                // Replace, Replace
                (
                    Some(Replace {
                        old: old1,
                        new: new1,
                    }),
                    Some(Replace {
                        old: old2,
                        new: new2,
                    }),
                ) => {
                    let op1 = next_op1.unwrap();
                    let op2 = next_op2.unwrap();

                    if op1.len_post() < op2.len_pre() {
                        change.push_op(
                            Replace {
                                old: old1,
                                new: (0, 0),
                            },
                            &self.buffer,
                        );
                        next_op1 = ops1.next();
                        next_op2 = Some(Replace {
                            old: (old2.0 + op1.len_post(), old2.1),
                            new: new2,
                        });
                    } else if op1.len_post() > op2.len_pre() {
                        change.push_op(
                            Replace {
                                old: old1,
                                new: (0, 0),
                            },
                            &self.buffer,
                        );
                        change.push_op(
                            Replace {
                                old: (0, 0),
                                new: new2,
                            },
                            &other.buffer,
                        );
                        next_op1 = Some(Replace {
                            old: (0, 0),
                            new: (new1.0 + op2.len_pre(), new1.1),
                        });
                        next_op2 = ops2.next();
                    } else {
                        change.push_op_separate_buf(
                            Replace {
                                old: old1,
                                new: new2,
                            },
                            &self.buffer,
                            &other.buffer,
                        );
                        next_op1 = ops1.next();
                        next_op2 = ops2.next();
                    }
                }
            }
        }

        change
    }

    /// Creates a change that is the inverse of this one (i.e. the "undo"
    /// of this change).
    pub fn inverse(&self) -> Change {
        let mut inv = Change::new();

        for op in self.ops.iter() {
            match op {
                Op::Retain(byte_count) => {
                    inv.retain(*byte_count);
                }
                Op::Replace {
                    old: old_range,
                    new: new_range,
                } => {
                    // Swap old and new to do the inverse.
                    let old = &self.buffer[new_range.0..new_range.1];
                    let new = &self.buffer[old_range.0..old_range.1];

                    inv.replace(old, new);
                }
            }
        }

        inv
    }

    //---------------------------------------------------------

    /// Pushes a retain op onto the change operations.
    #[inline(always)]
    fn retain(&mut self, byte_count: usize) {
        if let Some(Op::Retain(ref mut last_byte_count)) = self.ops.last_mut() {
            *last_byte_count += byte_count;
        } else if byte_count > 0 {
            self.ops.push(Op::Retain(byte_count));
        }
    }

    /// Pushes a replace op onto the change operations.
    #[inline(always)]
    fn replace(&mut self, old: &str, new: &str) {
        if !old.is_empty() || !new.is_empty() {
            let ti = self.buffer.len();
            self.buffer.push_str(old);
            let old_range = (ti, ti + old.len());

            let ti = self.buffer.len();
            self.buffer.push_str(new);
            let new_range = (ti, ti + new.len());

            self.ops.push(Op::Replace {
                old: old_range,
                new: new_range,
            });
        }
    }

    /// Pushes an op onto the change operations.
    ///
    /// - op: the operation.
    /// - buffer: the string buffer the op points into if it's a `Replace`.
    #[inline(always)]
    fn push_op(&mut self, op: Op, buffer: &str) {
        self.push_op_separate_buf(op, buffer, buffer);
    }

    /// Pushes an op onto the change operations.
    ///
    /// - op: the operation.
    /// - buf_old: the string buffer that `Replace { old }` points into.
    /// - buf_new: the string buffer that `Replace { new }` points into.
    #[inline(always)]
    fn push_op_separate_buf(&mut self, op: Op, buf_old: &str, buf_new: &str) {
        match op {
            Op::Retain(byte_count) => self.retain(byte_count),
            Op::Replace { old, new } => {
                self.replace(&buf_old[old.0..old.1], &buf_new[new.0..new.1])
            }
        }
    }
}

impl std::fmt::Debug for Change {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        f.write_str("Change {\n")?;
        f.write_str("  ops: [\n")?;
        for op in self.ops.iter() {
            match op {
                Op::Retain(byte_count) => f.write_str(&format!("    Retain({}),\n", byte_count))?,
                Op::Replace { old, new } => f.write_str(&format!(
                    "    Replace(\n      old({}): \"{}\",\n      new({}): \"{}\"\n    ),\n",
                    old.1 - old.0,
                    &self.buffer[old.0..old.1],
                    new.1 - new.0,
                    &self.buffer[new.0..new.1],
                ))?,
            }
        }
        f.write_str("  ],\n")?;
        f.write_str(&format!(
            "  buffer({}): \"{}\",\n",
            self.buffer.len(),
            &self.buffer
        ))?;
        f.write_str("}\n")?;

        Ok(())
    }
}

pub trait ApplyChange {
    /// Applies a Change.
    fn apply_change(&mut self, change: &Change);

    /// Applies the inverse of a Change.
    ///
    /// More-or-less an "undo", although for some types applying a change
    /// followed by applying it's inverse may not result in exactly the same
    /// state (cursor positions, for example).
    fn apply_change_inverse(&mut self, change: &Change);
}

impl ApplyChange for Rope {
    fn apply_change(&mut self, change: &Change) {
        let mut i = 0;
        for op in change.ops.iter() {
            match op {
                Op::Retain(byte_count) => {
                    i += byte_count;
                }
                Op::Replace { old, new } => {
                    let old = &change.buffer[old.0..old.1];
                    let new = &change.buffer[new.0..new.1];

                    if !old.is_empty() {
                        debug_assert_eq!(self.slice(i..(i + old.len())), old);
                        self.remove(i..(i + old.len()));
                    }
                    if !new.is_empty() {
                        self.insert(i, new);
                    }

                    i = i + new.len();
                }
            }
        }
        debug_assert!(i <= self.len());
    }

    fn apply_change_inverse(&mut self, change: &Change) {
        let mut i = 0;
        for op in change.ops.iter() {
            match op {
                Op::Retain(byte_count) => {
                    i += byte_count;
                }
                Op::Replace {
                    old: old_range,
                    new: new_range,
                } => {
                    // Swap old and new, to do the inverse.
                    let old = &change.buffer[new_range.0..new_range.1];
                    let new = &change.buffer[old_range.0..old_range.1];

                    if !old.is_empty() {
                        debug_assert_eq!(self.slice(i..(i + old.len())), old);
                        self.remove(i..(i + old.len()));
                    }
                    if !new.is_empty() {
                        self.insert(i, new);
                    }

                    i = i + new.len();
                }
            }
        }
        debug_assert!(i <= self.len());
    }
}
